;;;; AVSWORK -- AVSomat frameWORK -- commonlisp framework to integrate AVSOMAT
;;;;                                 and related tools and report on the web
;;;; Copyright (c) 2003-2004 Alexandru Dan Corlan MD PhD
;;;; 

;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2 of the License, or
;;;; (at your option) any later version.

;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.

;;;; You should have received a copy of the GNU General Public License
;;;; along with this program; if not, write to the Free Software
;;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


(require "AVSOBS" "/home/alexandru/proj/avsomat/src/avsobs.lisp")

(defun script (str &key (output-type :string))
  "Execute the STR string as a bash script in a new shell process."
  (cond
   ((eq output-type :string)
    (with-output-to-string (so)
       (with-input-from-string (si str)
	  (extensions:run-program "bash" () :input si :output so)
	  so)))
   ))

(defun fitsname-to-varname (fitsname)
  "Convert fits name nnccc-f-yyy.fits[.gz] into the name
   of the variable star, NN-CCC."
    (let* ((varn (subseq fitsname 0 (position #\- fitsname)))
	   (vl (- (length varn) 3))
	   )
      (intern (string-upcase
	       (concatenate 'string (subseq varn 0 vl) "-" (subseq varn vl))))
      )
    )

(defun already-reported (bull varstar)
  "Return T if VARSTAR reported in bull has already been reported to aavso."
  (dolist (re (get varstar 'aavso-rep))
    (if (equal (car re) (subseq bull 19 31))
	(return-from already-reported t)
;      (format *standard-output* "+~S+ +~S+~%" (car re) (subseq bull 19 31))
      )
    )
  nil
  )

(defun run-avsomat (fits obset obselt &key (tycho t)
		     (correct t) 
		     (rcp-dir *rcp2-dir*)
		     (obscode *default-obscode*)
		     )
  "Run avsomat to reduce the FITS file to an aavso report.
   OBSET is usually a 4-digit integer specifying the night of the observation
   (last digits of julian date). OBSELT is number of the observation of
   that object in that night.
   If :TYCHO, use tycho references instead of the aavso ones.
   Recipe files are taken from :RCP-DIR."

  ;; the start of the fits filename must identify the star
  (let* ((fitsn (if (pathnamep fits)
		    fits
		  (parse-namestring fits)))
	 (varname (fitsname-to-varname (pathname-name fitsn))))
    (if (not (probe-file (format nil "~A/~A.rcp" rcp-dir varname)))
	(return-from run-avsomat 'no-rcp-available))
    (if (string= "gz" (pathname-type fitsn))
	(script (format nil "gzip -cd ~A >~A/temp/~A.fits" fits *avsomat-home*
			varname))
      (script (format nil "cp ~A ~A/temp/~A.fits" fits *avsomat-home* varname))
      )
    (script (format nil "~A/bin/sex ~A/temp/~A.fits"
		    *avsomat-home*
		    *avsomat-home* varname
		    ))
    (when (not (probe-file (format nil "~A/~A" *report-dir* varname)))
      (script (format nil "mkdir ~A/~A/~A" *report-dir* obset varname)))
    (let ((bull
	   (script (format
		    nil
		    "~A/bin/avsomat ~A -N~A -b~A/~A/~A/~A.txt -P~A/~A/~A/~A.eps -Ctest.cat -c~A ~A/temp/~A.fits ~A/~A.rcp"
		    *avsomat-home*
		    (if tycho
			(if correct "-tj" "-t")
		      "")
		    obscode
		    ; txt report
		    *report-dir* obset varname obscode
		    ; eps:
		    *report-dir* obset varname obscode
		    ; comment:
		    (format nil "astro.corlan.net/obs/~A/~A/~A.html"
			    obset varname obselt)
		    ; fits:
		    *avsomat-home* varname
		    ; rcp:
		    rcp-dir varname))
	   ))
      (script 
       (format nil 
	 "gcx -n ~A/temp/~A.fits | convert -  -flip -comment \"~A\" -geometry 50%x50% ~A/~A/~A/~A.jpg"
	 *avsomat-home* varname
	    (format nil "
Sky image of variable star ~A
on the night of JD-245~A. Produced with GCX.
Copyright (c) 2004 R&D Corlan.
For details see http://astro.corlan.net ." varname obset)
	 *report-dir* obset varname obscode))
      (script (format nil "rm ~A/temp/~A.fits ; rm test.cat" *avsomat-home* varname))
      (if (already-reported bull varname)
	  (format *standard-output* "previously reported: "))
      (format *standard-output* "~A" bull)
      )
    )
  )


(defun do-obstar-obs (obset varname n obs)
  "Convert the results which have been placed in the directory DIR by
   avsomat into html+gif form, generating N.html for observer OBS."
  (with-open-file (fo (format nil "~A/~A/~A/~A.html" *report-dir* obset varname n)
		       :direction :output :if-exists :supersede)
    (format fo
       "<HTML><HEAD><TITLE>AVSOMAT observation on variable star ")
    (format fo " ~A (~A), JD245~A</TITLE></HEAD>~%"
       varname (full-name varname) obset obs)
    (format fo "<BODY BGCOLOR=\"#DDDDDD\">")
    (format fo "<TT>Observation bulletin for ")
    (format fo "<A HREF=\"http://astro.corlan.net/stars/~A/~A.html\">~A</A>,~%"
	    (constellation-name varname) varname (full-name varname))
    (format fo "<A HREF=\"http://astro.corlan.net/obs/~A/index.html\">JD 245~A</A>~%"
	    obset obset)
    (format fo "<BR><A HREF=\"http://astro.corlan.net\">[Home]</A>~%")
    (format fo "<A HREF=\"#comment\">[Comment on this observation]</A>~%")
    (format fo "<A HREF=\"http://www.aavso.org/cgi-bin/superstar2.pl?")
    (format fo "date1=~A&date2=~A&whocolor=3&who=~A&objectname=~A%20~A&grid=on&visual=on&fainter=on&ccdb=on&ccdv=on&ccdr=on&ccdi=on&ccdu=on\">"
	    (+ obset 2449970) (+ obset 2450030)
	    obs (let* ((sn (symbol-name varname)) (l (length sn)))
		  (subseq sn 0 (- l 4))) (constellation-name varname))
    (format fo "[AAVSO LightCurve, 60 days]</A>~%")
    (format fo "<A HREF=\"http://www.aavso.org/cgi-bin/superstar2.pl?")
    (format fo "date1=~A&date2=~A&whocolor=3&who=~A&objectname=~A%20~A&grid=on&visual=on&fainter=on&ccdb=on&ccdv=on&ccdr=on&ccdi=on&ccdu=on\">"
	    (+ obset 2449900) (+ obset 2450100)
	    obs (let* ((sn (symbol-name varname)) (l (length sn)))
		  (subseq sn 0 (- l 4))) (constellation-name varname))
    (format fo "[200 days]</A></TT><BR>~%")
    (format fo "<TABLE><TR><TD><IMG SRC=\"~A.jpg\" ALT=\"~A\"></TD>~%" obs
	    (format nil "Sky image of variable star ~A (~A) on the night of JD245~A."
		    varname (full-name varname) obset))
    (format fo "<TD><IMG SRC=\"~A-~A-~A.gif\" ALT=\"~A\"></TD></TR></TABLE>~%"
	    obset varname obs 
	    (format nil 
		"Identification sketch for variable star ~A (~A) on the night of JD245~A."
		varname (full-name varname) obset))
    (format fo "~%<P><PRE>")
    (with-open-file (fi (format nil "~A/~A/~A/~A.txt" *report-dir* obset varname obs))
		    (block nil
		      (do () (nil nil)
			(let ((la (read-line fi nil)))
			  (when (not la) (return))
			  (format fo "~A~%" la))))
		    )
    (format fo "</PRE>~%")
    (format fo "<A NAME=\"comment\">~%")
    (format fo "<P><FORM METHOD=POST ACTION=\"http://astro.corlan.net/cgi-bin/mailim\">~%")
    (format fo "<B>Send us a comment</B> on this observation,~%")
    (format fo "or just press [submit] to indicate~%")
    (format fo "that you think we should check it more carefully.<BR>~%")
    (format fo "<TEXTAREA NAME=CMT ROWS=4 COLS=60></TEXTAREA><P>~%")
    (format fo "Your name/e-mail (optional).<BR>~%")
    (format fo "<INPUT SIZE=60 NAME=EMAIL>~%")
    (format fo "<INPUT TYPE=SUBMIT><P>~%")
    (format fo "We will not use~%")
    (format fo "your e-mail for any other purpose except answering to your comment, nor store it in a database.<BR>~%")
    (format fo "</FORM>~%~%")
    (format fo "<P><HR><P>~%")
    (format fo "<TABLE WIDTH=100%><TR><TD ALIGN=RIGHT><FONT SIZE=-1>page produced with ")
    (format fo "<A HREF=\"http://astro.corlan.net/avsomat\">avsomat 3052</A></TD></TR>~%")
    (format fo "</TABLE>~%</BODY>~%</HTML>~%~%")
    )
  (if (probe-file (format nil "~A/~A/~A/~A.eps" *report-dir* obset varname obs))
      (script
       (concatenate 'string
	 (format nil "cd ~A/~A/~A~%" *report-dir* obset varname)
	 (format nil "cp ~A.eps temp.eps~%" obs)
	 (format nil "echo showpage quit >>temp.eps~%")
	 (format nil "gs -sDEVICE=pcx256 -sOutputFile=temp.pcx -r110 temp.eps </dev/null~%")
	 (format nil "convert temp.pcx -crop 0x0 ~A-~A-~A.gif~%"
		 obset varname obs)
	 (format nil "convert ~A-~A-~A.gif -crop 546x368+0+162 -comment \"~A\" ~A-~A-~A.gif~%"
		 obset varname obs
		 (format nil "
Identification sketch of variable star ~A
on the night of JD-245~A. Produced with AVSOMAT.
Copyright (c) 2004 R&D Corlan.
For details see http://astro.corlan.net." varname obset)
		 obset varname obs)
	 (format nil "rm temp.pcx temp.eps~%")
	 ))
    )
  )


(defun avsomat-dir (dir obset &key (tycho t)
			(correct t) (rcp-dir *rcp2-dir*)
			(start-with nil) (stop-after nil))
  "Run AVSOMAT with the parameters above on all fits files in the directory."
  (let ((dirl (directory (concatenate 'string dir "/*")))
	(prepare-to-stop nil)
	)
    (dolist (fn dirl)
      (with-simple-restart (nil "Ignore star, go to the next")
	(format *standard-output* (pathname-name fn))
	(when (and
	       (or (search "001" (pathname-name fn))
		   (search "002" (pathname-name fn)))
	       (or (string= "fits" (pathname-type fn))
		   (search "fits" (pathname-name fn)))
	       (or (search "-v-" (pathname-name fn))
		   (search "-clear-" (pathname-name fn)))
	       (> (position #\- (pathname-name fn)) 3)
	       )
	  (let ((obsc (if (search "002" (pathname-name fn)) 'CXR 'CUA))
		(obsi (if (search "001" (pathname-name fn)) 1 2))
		(varname (fitsname-to-varname (pathname-name fn)))
		)
	    (if (eq start-with varname)
		(setf start-with nil))
	    (if (and prepare-to-stop (not (eq stop-after varname)))
		(return-from avsomat-dir))
	    (if (not (probe-file (format nil "~A/~A/~A" *report-dir* obset varname)))
		(script (format nil "mkdir ~A/~A/~A" *report-dir* obset varname)))
	    (when (not start-with)
	      (run-avsomat fn obset obsi
			   :tycho tycho
			   :correct correct
			   :rcp-dir rcp-dir
			   :obscode obsc)
	      (when (or (probe-file
			 (format nil "~A/~A/~A/CUA.txt" *report-dir* obset varname))
			(probe-file
			 (format nil "~A/~A/~A/CUA.txt" *report-dir* obset varname)))
		(do-obstar-obs obset varname obsi obsc)
		)
	      )
	    (if (eq stop-after varname)
		(setf prepare-to-stop t))
	    )
	  )
	)
      )
    )
  )


(defun extract-magnitude (text-fn)
  "Extract the magnitude indication from an avsomat report, to be found
   in file text-fn."
  (if (probe-file text-fn)
    (with-open-file (fi text-fn)
		    (unless (read-line fi nil nil)
		      (return-from extract-magnitude '--))
		    (unless (read-line fi nil  nil)
		      (return-from extract-magnitude '--))
		    (let ((li (read-line fi nil nil)))
		      (unless li (return-from extract-magnitude '--))
		      (subseq li 30 35))
		    )
    '-)
  )

(defun chart-status (vs)
  (let ((varstar (if (symbolp vs)
		     vs (intern vs))))
    (if (get varstar 'chart-stat)
	(get varstar 'chart-stat)
      'unchecked)
    )
  )

(defun do-obstar-dir (obset &key (remake nil))
  "Generate the index file for the obset directory."
  (with-open-file (fo (format nil "~A/~A/index.html"  *report-dir* obset)
		      :direction :output :if-exists :supersede)
    (let ((names (directory (format nil "~A/~A/*" *report-dir* obset))))
      (format fo
	      "<HTML><HEAD><TITLE>AVSOMAT observations, 245-~A</TITLE></HEAD>~%"
	      obset)
      (format fo "<BODY BGCOLOR=\"#DDDDDD\">~%<PRE>")
      (format fo "Observations JD-245-~A</PRE><P>~%~%" obset)
      (format fo "<TABLE>~%")
      (dolist (nm names)
	(let* ((varstar (car (last (pathname-directory nm))))
	       (varcua (format nil "~A/~A/~A/CUA.txt" *report-dir* obset varstar))
	       (varcxr (format nil "~A/~A/~A/CXR.txt" *report-dir* obset varstar))
	       )
	  (format *standard-output* "~A~%" varcua)
	  (when (probe-file varcua)
	    (format *standard-output* "~A ... " varstar)
	    (when remake
		(do-obstar-obs obset (intern varstar) 1 'cua)
		(when (probe-file varcxr)
		  (do-obstar-obs obset (intern varstar) 2 'cxr)
		  )
		)
	    (format fo "<TR><TD>~A</TD><TD><A HREF=~A/1.html>~A</A></TD>"
		    varstar varstar (extract-magnitude varcua))
	    (format fo "<TD><A HREF=~A/2.html>~A</A></TD><TD>~A</TD></TR>~%"
		     varstar (extract-magnitude varcxr)
		    (chart-status varstar))
	    (format *standard-output* "done~%")
	    )
	  ))
	(format fo "</TABLE><P><A HREF=\"/obs/index.html\">All observation sets</A>~%</BODY>~%</HTML>~%~%")
      )
    )
  )

(defun redo-all-obstar-dirs (&optional (dirlist '(2791 2819 2833 2840 2875 2903 2910
						  2994 3022 3042 3045 3057 3065 3072
						  3093 3189 3236 3237)))
  (mapcar #'(lambda (x) (do-obstar-dir x :remake t))
	  dirlist))

(defun obs-volumes ()
  "Return the volumes available at the reports site."
  (set-difference (mapcar #'(lambda (x) (intern (car (last (pathname-directory x)))))
			  (directory (format nil "~A/*" *report-dir*))
			  )
		  '(|obs|))
  )

(defun stars-in-obs-volume (volume)
  "Return a list of stars which are reported in the specified VOLUME
   of observations from the reports site."
  (set-difference
   (mapcar #'(lambda (x) (intern (car (last (pathname-directory x)))))
	   (directory (format nil "~A/~A/*" *report-dir* volume))
	   )
   (list volume))
  )

(defun obs-on-site ()
  "Return an assoc list of volumes and stars in each volume which
   have avsomat bulletins on this site."
  (mapcar #'(lambda (x) (cons x (stars-in-obs-volume x)))
	  (obs-volumes)))

(defun count-obs (varstar obs-on-site)
  "Number of observations of the variable star in the obs-on-site structure"
  (let ((r 0))
    (mapc #'(lambda (x) (setf r (+ r (count varstar x))))
	  obs-on-site)
    r)
  )

(defvar *star-homes* "/home/alexandru/proj/web/astro.corlan.net/htdocs/stars"
  "Directory where monografies of stars are kept")

(defun do-varstar-dir (varstar volist
		       &optional (homedir *star-homes*))
  "Print all data avswork knows about the star in its proper directory."
  (let ((starfn (format nil "~A/~A/~A.html" homedir (constellation-name varstar) varstar)))
    (ensure-directories-exist starfn)
    (with-open-file (fo starfn :direction :output :if-exists :supersede)
       (format fo "<HTML><HEAD><TITLE>Variable star ~A (~A)</TITLE></HEAD>~%"
	       (full-name varstar) varstar)
       (format fo "<BODY>~%<FONT SIZE=+2>~A (~A)</FONT><BR>~%"
	       (full-name varstar)
	       (get varstar 'aavsoid))
       (format fo "<A HREF=\"http://astro.corlan.net\">[Home]</A> ")
       (format fo "<A HREF=\"http://astro.corlan.net/stars/index.html\">[AllStars]</A> ")
       (format fo "<A HREF=\"http://astro.corlan.net/stars/~A/index.html\">[~A]</A> "
	       (constellation-name varstar) (constellation-name varstar))
       (format fo "<A HREF=\"http://astro.corlan.net/obs/index.html\">[AllObs]</A> ")
       (format fo "<A HREF=\"http://www.aavso.org/cgi-bin/superstar2.pl?")
       (format fo "objectname=~A%20~A&visual=on&fainter=on&ccdb=on&ccdv=on&ccdr=on&ccdi=on&ccdu=on\">[AAVSO LightCurve, 1yr]</A>~%"
	       (let* ((sn (symbol-name varstar)) (l (length sn)))
		 (subseq sn 0 (- l 4))) (constellation-name varstar))
       (format fo "<A HREF=\"http://www.aavso.org/cgi-bin/superstar2.pl?")
       (format fo
	  "objectname=~A%20~A&date1=2451250&visual=on&fainter=on&ccdb=on&ccdv=on&ccdr=on&ccdi=on&ccdu=on\">[AAVSO LightCurve, 5yr]</A></TT><BR>~%"
	       (let* ((sn (symbol-name varstar)) (l (length sn)))
		 (subseq sn 0 (- l 4))) (constellation-name varstar))
       (format fo
   "<P><B><A HREF=\"http://www.sai.msu.su/groups/cluster/gcvs/gcvs\">GCVS</A> data:</B>")
       (format fo "<BR>Epoch 2000: ra = ~A, dec = ~A<BR>~%"
	       (get varstar 'ra) (get varstar 'dec))
       (format fo "Variability type: ~A<BR>~%" (get varstar 'vartype))
       (format fo "Spectral type: ~A<BR>~%" (get varstar 'spectype))
       (format fo "<P><B>GCVS remarks:</B><BR>~%")
       (format fo "~A" (get varstar 'gcvs-remark))
       (format fo "<P><TABLE><TR><TD><B>Search (~A):</B></TD>~%<TD>" varstar)
       (format fo "<A HREF=\"http://www.google.com/search?q=~A\">Google</A>~%"
	       varstar)
       (format fo "<A HREF=\"http://www.scirus.com/srsapp/search?q=~A\">Scirus</A>~%"
	       varstar)
       (format fo
     "<A HREF=\"http://www.google.com/search?q=~A%20site%3Awww.aavso.org\">AAVSO</A>~%"
	       varstar)
       (format fo "</TD></TR>~%<TR><TD><B>Search (~A):</B></TD>" (full-name varstar))
       (format fo "<TD><A HREF=\"http://www.google.com/search?q=%22~A%22\">Google</A>~%"
	       (full-cgi-name varstar))
       (format fo "<A HREF=\"http://www.scirus.com/srsapp/search?q=%22~A%22\">Scirus</A>~%"
	       (full-cgi-name varstar))
       (format fo
     "<A HREF=\"http://www.google.com/search?q=%22~A%22%20site%3Awww.aavso.org\">AAVSO</A>~%"
	       (full-cgi-name varstar))
       (format fo "</TD></TR></TABLE>~%")
       (format fo "<P>")
       (when (get varstar 'links)
	 (format fo "<B>Relevant links:</B><BR>~%<UL>~%")
	 (dolist (l (get varstar 'links))
	   (format fo "<li><A HREF=\"~A\">~A</A>" (car l) (cadr l))
	   (if (cddr l)
	       (format fo "<BR>~A~%" (caddr l)))
	   )
	 (format fo "</UL>~%~%")
	 )
       (format fo "<B>Chart status:</B> ~A<P>~%" (chart-status varstar))
       (format fo "<B>Observations on this site:</B><BR>~%")
       (dolist (v volist)
	 (format fo
	   "245~A <A HREF=/obs/~A/~A/1.html>~A</A> <A HREF=/obs/~A/~A/2.html>~A</A><P>~%"
	   v v varstar
	   (extract-magnitude (format nil "~A/~A/~A/CUA.txt" *report-dir* v varstar))
	   v varstar
	   (extract-magnitude (format nil "~A/~A/~A/CXR.txt" *report-dir* v varstar))
	   ))
       (format fo "<A NAME=\"comment\">~%")
       (format fo "<P><FORM METHOD=POST ACTION=\"http://astro.corlan.net/cgi-bin/mailim\">~%")
       (format fo "<B>Send us a comment,</B> a reference or a link about this star,~%")
       (format fo "that you think we should add on this page.<BR>~%")
       (format fo "<TEXTAREA NAME=CMT ROWS=4 COLS=60></TEXTAREA><P>~%")
       (format fo "Your name/e-mail (optional).<BR>~%")
       (format fo "<INPUT SIZE=60 NAME=EMAIL>~%")
       (format fo "<INPUT TYPE=SUBMIT><P>~%")
       (format fo "We will not use~%")
       (format fo "your e-mail for any other purpose, nor store it in a database.<BR>~%")
       (format fo "</FORM>~%~%")
       (format fo "<P><HR><TABLE WIDTH=100%><TR><TD ALIGN=RIGHT>")
       (format fo "<FONT SIZE=-2>Page generated by ")
       (format fo "<A HREF=\"http://astro.corlan.net/avsomat\">Avsomat</A></FONT>")
       (format fo "</TD></TR></TABLE>~%")
       (format fo "</BODY></HTML>~%")
       
       )
    ))


(defun do-all-varstar-dirs (&optional (starlist *vs*))
  "Make all directories and star files for the stars in the STARLIST."
  (let* ((oos (obs-on-site)))
    (mapc #'(lambda (x)
	      (do-varstar-dir x 
		 (mapcan #'(lambda (y)
			     (if (member x (cdr y)) (list (car y)) nil))
			 oos))
	      
	      )
	  starlist)
    (let ((constels (directory (format nil "~A/*" *star-homes*)))
	  (totobs 0))
      (with-open-file (fo (format nil "~A/index.html" *star-homes*)
			  :direction :output :if-exists :supersede)
        (format fo "<HTML><HEAD><TITLE>Available constellations</TITLE></HEAD>~%")
	(format fo "<BODY BGCOLOR=\"#DDDDDD\">~%")
	(format fo
		"<CENTER><FONT SIZE=+2>Variable stars with data on our site</FONT><BR>~%")
	(format fo "select a constellation<P>~%")
	(format fo "<TABLE WIDTH=60%><TR>~%")
	(let ((colj 0)
	      (nobs 0))
	  (dolist (co constels)
	    (let* ((coname (car (last (pathname-directory co))))
		   (condir (directory
			    (format nil "~A/~A/*" *star-homes* coname)))
		   )	      
	      (when (not (string= coname "stars"))
		(format fo "<TD ALIGN=RIGHT><A HREF=~A>~A</A>~%" coname coname)
		(setf nobs 0)
		(with-open-file (fo2 (format nil "~A/~A/index.html" *star-homes* coname)
				     :direction :output :if-exists :supersede)
				(format fo2
				 "<HTML><HEAD><TITLE>Variable stars in ~A</TITLE>" coname)
				(format fo2 "</HEAD>~%<BODY BGCOLOR=\"#DDDDDD\">~%")
				(format fo2 "<CENTER><FONT SIZE=+3>")
				(format fo2 "Variable stars in ~A</FONT><BR>~%" coname)
				(format fo2 "<TABLE><TR>~%")
				(let ((coli 0) (nobsl 0))
				  (dolist (st condir)
				    (let ((starf (pathname-name st)))
				      (when (not (string= starf "index"))
					(setf nobsl (count-obs (intern starf) oos))
					(incf nobs nobsl)
					(format fo2
                                         "<TD ALIGN=RIGHT><A HREF=~A.html>~A</A>[~A]</TD>~%"
					 starf starf
					 nobsl))
				      )
				    (incf coli)
				    (if (zerop (mod coli 6))
					(format fo2 "</TR><TR>"))
				    )
				  )
				(format fo2 "</TR></TABLE>~%")
				(format fo2 "<A HREF=http://astro.corlan.net>")
				(format fo2 "<IMG SRC=/btns/Home.gif BORDER=0></A>")
				(format fo2 "<A HREF=http://astro.corlan.net/stars>")
				(format fo2 "<IMG SRC=/btns/Stars.gif BORDER=0></A>")
				(format fo2 "<A HREF=http://astro.corlan.net/obs>")
				(format fo2 "<IMG SRC=/btns/Obs.gif BORDER=0></A>")
				(format fo2 "<HR><TABLE WIDTH=100%><TR><TD ALIGN=RIGHT>")
				(format fo2 "<FONT SIZE=-2>Page generated by ")
				(format fo2
			 "<A HREF=\"http://astro.corlan.net/avsomat\">Avsomat</A></FONT>")
				(format fo2 "</TD></TR></TABLE>~%")
				(format fo2 "</CENTER></BODY></HTML>~%")
				)
		(format fo "~A</TD>~%" nobs)
		(incf totobs nobs)
		(incf colj)
		(when (zerop (mod colj 8))
		  (format fo "</TR>~%<TR>~%"))

		)
	      )
	    )
	  )
	(format fo
	   "</TR></TABLE>(~A pairs of observations in total)</CENTER>~%" totobs)
	(format fo "<HR><TABLE WIDTH=100%><TR><TD ALIGN=RIGHT>")
	(format fo "<FONT SIZE=-2>Page generated by ")
	(format fo "<A HREF=\"http://astro.corlan.net/avsomat\">Avsomat</A></FONT>")
	(format fo "</TD></TR></TABLE>~%")
	(format fo "</BODY></HTML>~%")
	)
      )
    )
  )



