-- DIASKYS -- Distributed Involuntary Amateur Sky Survey
-- Copyright (c) 2004 Alexandru Dan Corlan MD PhD (http://dan.corlan.net)

-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-- Diaskys uses a database with the positions
-- and timings of pictures of the sky that were presumably
-- taken by an amateur astronomer with his CCD camera; it is
-- given either: (1) a fixed position on the sky and reports all the
-- images in the database containing that position;
-- (2) two position and two instants in time and it reports
-- all the images on which an object travelling with constant
-- speed between the two points at the two instants would be caught

-- Thus you can find if the particular astronomer who runs
-- diaskys on his site has images on which an object (star or
-- moving object) you are interested in should be found, and
-- you can tell him to look at those images or send them to you

-- Data can be requested through either a web form or an URI scheme.
-- In the future they might be requested over SOAP, so it
-- will be simple to ask all known servers about a position
-- of interest.



package Diaskys_Data is

   type Plate is record
      Ra_Min, Ra_Max, Dec_Min, Dec_Max, Jd: Long_Float;
      Fn: String(1..100);
      Fn_Fill: Integer:= 0;
   end record;

   Max_Plates: Integer:= 100_000;

   type Db is array(0..Max_Plates) of Plate;

   Pdb: Db;
   Pdb_Fill: Integer:= 0;

end Diaskys_Data;

with Ada.Text_IO; use Ada.Text_Io;

with AWS.Response;
with AWS.Server;
with AWS.Status;
with AWS.Default;
with Aws.Parameters;
with Aws.Mime;

with Unisig; use Unisig;
with Ada.Exceptions; use Ada.Exceptions;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Diaskys_Data; use Diaskys_Data;
with Ada.Numerics.Long_Elementary_functions; use Ada.Numerics.Long_Elementary_functions;

procedure Diaskysd is

   WS : AWS.Server.HTTP;
   Dbf: File_Type;
   Dbl: String(1..200);
   Linel: Integer;

--   package Math is new Ada.Numerics.Generic_Elementary_Functions(Long_Float);

   function Diaskys_Server (Request : in AWS.Status.Data)
                            return AWS.Response.Data is

      P: Aws.Parameters.List:= Aws.Status.Parameters(Request);
      Uri: String:= Aws.Status.Uri(Request);
      Npa: Integer:= Aws.Parameters.Count(P);
      O: Unbounded_String:= To_Unbounded_String("");
      Not_Given: constant Long_Float:= -999.0;
      Ra0, Jd0, Dec0, Ra1, Jd1, Dec1: Long_Float:= Not_Given;
      Apded: array(0..Pdb_Fill) of Boolean:= (others => False);

      procedure Append_Images_At(Ra, Dec, Jd, Jdur: Long_Float:= Not_Given) is

      begin
         if (Ra/=Not_Given and Dec/=Not_Given) then
            for I in 0..(Pdb_Fill-1) loop
               if not Apded(I) and then
                 (Ra>Pdb(I).Ra_Min and Ra<Pdb(I).Ra_Max and
                  Dec>Pdb(I).Dec_Min and Dec<Pdb(I).Dec_Max and
                  ((Jd=Not_Given and Jdur=Not_Given) or
                   (Jd<=Pdb(I).Jd and (Jd+Jdur)>=Pdb(I).Jd))) then
                  Append(O, Pdb(I).Fn(1..Pdb(I).Fn_Fill) & Ascii.Lf);
                  Apded(I) := True;
               end if;
            end loop;
         end if;
      end Append_Images_At;

   begin

      if Uri="/" then
         return
           Aws.Response.Build(
          "text/html",
          "<HTML><HEAD><TITLE>Distributed Involuntary Amateur SKY Survey</TITLE></HEAD>"
          & Ascii.Lf & "<BODY BGCOLOR=""#DDDDDD"">" & Ascii.Lf &
          "<CENTER><FONT SIZE=+2>DIASKYS</FONT><BR>" & Ascii.Lf &
          "Distributed Involuntary Amateur SKY Survey<P>" & Ascii.Lf &
     --          "<FORM METHOD=GET ACTION=""http://astro.corlan.net:20206"">" & Ascii.Lf &
          "<FORM METHOD=GET ACTION=""/search"">" & Ascii.Lf &
          "<TABLE><TR><TD></TD><TD>RA (deg)</TD><TD>DEC (deg)</TD>" &
          "<TD>Julian date</TD></TR>" & Ascii.Lf & "<TR><TD>First point</TD>" & Ascii.Lf &
          "<TD><INPUT NAME=RA0 SIZE=10></TD>" & Ascii.Lf &
          "<TD><INPUT NAME=DEC0 SIZE=10></TD>" & Ascii.Lf &
          "<TD><INPUT NAME=JD0 SIZE=20></TD><TR>" & Ascii.Lf &
          "<TR><TD>Last point</TD>" & Ascii.Lf &
          "<TD><INPUT NAME=RA1 SIZE=10></TD>" & Ascii.Lf &
          "<TD><INPUT NAME=DEC1 SIZE=10></TD>" & Ascii.Lf &
          "<TD><INPUT NAME=JD1 SIZE=20></TD><TR>" & Ascii.Lf &
          "</TABLE><P><INPUT TYPE=SUBMIT></CENTER></FORM>" & Ascii.Lf &
          "</BODY></HTML>" & Ascii.Lf & Ascii.lf
          );
      else
         for Zz in 1..Npa loop
            declare
               Nm: String:= Aws.Parameters.Get_Name(P,Zz);
               Vl: String:= Aws.Parameters.Get_Value(P,Zz);
            begin
               if Vl'Length>0 then
                  if Nm="RA0" then
                     Ra0 := Long_Float'Value(Vl);
                  elsif Nm="DEC0" then
                     Dec0 := Long_Float'Value(Vl);
                  elsif Nm="JD0" then
                     JD0 := Long_Float'Value(Vl);
                  elsif Nm="RA1" then
                     Ra1 := Long_Float'Value(Vl);
                  elsif Nm="DEC1" then
                     Dec1 := Long_Float'Value(Vl);
                  elsif Nm="JD1" then
                     JD1 := Long_Float'Value(Vl);
                  end if;
               end if;
            end;
         end loop;
      end if;

      -- just do it for ra0, dec0, ignore all JD & stuff

      Append(O, "<HTML><HEAD><TITLE>Files available at astro.corlan.net for ra=" &
             Long_Float'Image(Ra0) & ", dec=" & Long_Float'Image(Dec0));
      Append(O, "</TITLE></HEAD>" & Ascii.Lf);
      Append(O, "<BODY>" & Ascii.Lf);
      Append(O, "Files available at astro.corlan.net for ra=" &
             Long_Float'Image(Ra0) & ", dec=" & Long_Float'Image(Dec0) &
             "<P>" & Ascii.Lf);
      Append(O, "<PRE>" & Ascii.Lf);

      if (Ra0/=Not_Given and Dec0/=Not_Given and Jd0=Not_Given
          and Ra1=Not_Given and Dec1=Not_Given and Jd1=Not_Given) then
         Append_Images_At(Ra0, Dec0);
      end if;

      if (Ra0/=Not_Given and Dec0/=Not_Given and Jd0/=Not_Given
          and Ra1/=Not_Given and Dec1/=Not_Given and Jd1/=Not_Given) then
         declare
            Crt_Jd: Long_Float:= Jd0;
            Dist: Long_Float:= Sqrt((Ra1-Ra0)*(Ra1-Ra0)+(Dec1-Dec0)*(Dec1-Dec0));
            Nsteps: Integer:= Integer(Dist/0.01);
            Ddist: Long_Float := Dist / Long_Float(Nsteps);
            Jdist: Long_Float := (Jd1 - Jd0) / Long_Float(Nsteps);
         begin
            Append_Images_At(Ra0, Dec0, Jd0, Jdist);
            Append_Images_At(Ra1, Dec1, Jd1-Jdist, Jdist);
            for I in 1..Nsteps loop
               Append_Images_At(Ra0+Long_Float(I)*(Ra1-Ra0)/Long_Float(Nsteps),
                                Dec0+Long_Float(I)*(Dec1-Dec0)/Long_Float(Nsteps),
                                Jd0+Jdist*Long_Float(I)/Long_Float(Nsteps), Jdist);
            end loop;
         end;
      end if;


      Append(O, "</PRE></BODY></HTML>" & Ascii.Lf);
      return Aws.Response.Build("text/html", To_String(O));

   exception
      when Error: others =>
         Put_Line("Some error in the server, sorry" & Ascii.Lf &
                  Exception_Name(Error) & Ascii.Lf &
                  Exception_Message(Error) & Ascii.Lf &
                  Exception_Information(Error) & Ascii.Lf );
         return Aws.Response.Build("text/html",
                                   "<P> Sorry, internal server error" & Ascii.Lf &
                                   Exception_Name(Error) & Ascii.Lf &
                                   Exception_Message(Error) & Ascii.Lf &
                                   Exception_Information(Error) & Ascii.Lf);
   end Diaskys_Server;

begin

   Open(Dbf, In_File, "/web/astro.corlan.net/data/diaskys.data");
   Pdb_Fill := 0; -- was 0 anyway :)
   loop
      exit when End_Of_File(Dbf);
      Get_Line(Dbf, Dbl, Linel);
      if Linel>75 then
         Pdb(Pdb_Fill).Ra_Min := Long_Float'Value(Dbl(1..12));
         Pdb(Pdb_Fill).Ra_Max := Long_Float'Value(Dbl(14..25));
         Pdb(Pdb_Fill).Dec_Min := Long_Float'Value(Dbl(27..38));
         Pdb(Pdb_Fill).Dec_Max := Long_Float'Value(Dbl(40..51));
         Pdb(Pdb_Fill).Jd := Long_Float'Value(Dbl(53..64));
         Pdb(Pdb_Fill).Fn(1..Linel-74) := "" & Dbl(75..Linel);
         Pdb(Pdb_Fill).Fn_Fill := Linel-74;
         Pdb_Fill := Pdb_Fill + 1;
      end if;
   end loop;
   Close(Dbf);

   AWS.Server.Start (WS, "Diaskys",
                     Max_Connection => 10,
                     Port => 20206,
                     Callback       => Diaskys_Server'Unrestricted_Access);
   loop
      delay 1.0;
      if Termination_Flag.Istermi then
         exit;
      end if;
   end loop;

   Put_Line("OK, Diaskys leaving gracefully.");

   AWS.Server.Shutdown (WS);
end Diaskysd;


-- the following utilities allow stoping the server graciously with kill -3

with Ada.Interrupts; use Ada.Interrupts;
with Ada.Interrupts.Names; use Ada.Interrupts.Names;

package Unisig is

   protected Termination_Flag is

      procedure Termi;
      pragma Interrupt_Handler(Termi);
      pragma Attach_Handler(Termi,SIGQUIT);

      function Istermi return Boolean;

   private

      Stopit: Boolean:= False;

   end Termination_Flag;

end Unisig;


package body Unisig is

   protected body Termination_Flag is

      procedure Termi is

      begin
         Stopit := True;
      end Termi;

      function Istermi return Boolean is

      begin
         return Stopit;
      end Istermi;

   end Termination_Flag;

end Unisig;

