// Copyright (c) 2003 Alexandru Dan Corlan (http://dan.corlan.net)
// Matching two sets of stars, presumably with the same pattern

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>

#define PI 3.1415926

static float maxdist=1000.0;

typedef struct {
  double a, b, c;
  double perim;   // a+b+c of course, but saves two aditions per triangle pair
  double al, be, ga;
  int sa, sb, sc; // indices of the stars making up the triangle
  int useful; // 0 if a, b or c were found to be above maxdist
} TRI;

typedef struct {
  double x, y; // normally in pixels, converted with degperpix for catalog
  char id[15];
  float mag;
  int elt;
} STAR;

typedef struct {
  int ntri, nstar; // sizes
  int ftri, fstar; // fills
  TRI *t;
  STAR *s;
} PIC;


static STAR *cat; // set of stars from the catalogue, must be sorted in decreasing magnitude
static int catsz;

static int nis;
static int full_nis;

static int handsol[1000];
static int autosol[1000];
static int ref_solution;

static PIC pic;

static int list_final_solution_f = 0;
static int list_repeated_solutions_f = 0;
static int list_triangle_matches_f = 0;
static int silent_f = 0;

static char *err = NULL;

static error(char *msg) {
  err = msg;
}

// triangle match solutions 

typedef struct {
  short pa, pb, pc, ca, cb, cc;
  float omega, dx0, dy0, dx1, dy1, dmag, merr;
  short repeated;
  short like;
} TRISOL; // triangle solution

#define TSMAX 10000

static TRISOL ts[TSMAX];
static int tsfill = 0;

static int compute_ntri(int nis) {  //  (nis,3)

  int r;

  r = nis * (nis-1) * (nis-2) / 6;
  return r;
}

static int selective_compute_ntri(int nis) {  //  (nis,3)

  int r;

  r = nis * 10;
  return r;
}

static double dist(double x0, double y0, double x1, double y1) {
  return sqrt((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0));
}

static double unghi(double x0, double y0, double x1, double y1, double x2, double y2) {
  double a, b, c;

  a = dist(x0,y0,x1,y1);
  b = dist(x0,y0,x2,y2);
  c = dist(x1,y1,x2,y2);
  return acos((a*a - c*c + b*b) / (2*a*b));
}

// the following transform coordinates of the chart into the picture
// and are produced by the compute_transform procedure

static double transx, transy;  // to be added to the chart coordinates
static double omega2;          // to rotate the chart coordinates by
static double orgx, orgy;      // to rotate the chart coordinates around
                        // (to be substracted before substraction)
static double magx, magy;      // to multiply the chart coordinates with
static double omega1;          // to rotate the chart coordinates back with
static double atr=1.0;
static double btr=-0.14;
static double ctr=0.0;
static double dtr=0.93;        // coefficients for linear transform of coordinates
                               // after rotation; unit transform is specified as default

static void chart_to_pic(double x, double y, double catfact, double *newx, double *newy,
			 int with_ltr);

static void compute_linear_transform(int pb, int pc, int cb, int cc) {

  double x1, x2, y1, y2, x1p, x2p, y1p, y2p;

  x1p = pic.s[pb].x - orgx;
  y1p = pic.s[pb].y - orgy;
  x2p = pic.s[pc].x - orgx;
  y2p = pic.s[pc].y - orgy;
  chart_to_pic(cat[cb].x,cat[cb].y,1.0,&x1,&y1,0);
  chart_to_pic(cat[cc].x,cat[cc].y,1.0,&x2,&y2,0);
  x1 = x1 - orgx;
  x2 = x2 - orgx;
  y1 = y1 - orgy;
  y2 = y2 - orgy;

  atr = (x1p * y2 - y1 * x2p) / (x1 * y2 - y1 * x2);
  btr = (x1p * x2 - x1 * x2p) / (y1 * x2 - y2 * x1);
  ctr = (y1p * y2 - y1 * y2p) / (x1 * y2 - y1 * x2);
  dtr = (y1p * x2 - y2p * x1) / (y1 * x2 - y2 * x1);
}

static double invatr = 1.0;
static double invbtr = 0.0;
static double invctr = 0.0;
static double invdtr = 1.0;

static void compute_inverse_linear_transform_coefficients() {

  double numi;

  numi = atr * dtr - btr * ctr;
  invatr =  dtr / numi;
  invbtr = -btr / numi;
  invctr = -ctr / numi;
  invdtr =  atr / numi;
}

static void apply_inverse_linear_transform(STAR *s) {

  double xx, yy;

  xx = invatr * s->x + invbtr * s->y;
  yy = invctr * s->x + invdtr * s->y;

  s->x = xx;
  s->y = yy;
}

//#define PI 3.1415927

static void compute_transform(int pa, int pb, int pc, int ca, int cb, int cc) {
  float cax, cay, cbx, cby, ccx, ccy;
  float pax, pay, pbx, pby, pcx, pcy;
  float carx, cary, cbrx, cbry, ccrx, ccry;
  float parx, pary, pbrx, pbry, pcrx, pcry;
  float omegac;

  orgx = pic.s[pa].x;
  orgy = pic.s[pa].y;
  transx = pic.s[pa].x - cat[ca].x;
  transy = pic.s[pa].y - cat[ca].y;
  // below, may need to check that the x difference is not 0
  omega1 = atan2(pic.s[pb].y - pic.s[pa].y, pic.s[pb].x - pic.s[pa].x);
  if (omega1<0)
    omega1 = 2 * PI + omega1;
  omega2 = atan2(cat[cb].y - cat[ca].y, cat[cb].x - cat[ca].x);
  if (omega2<0)
    omega2 = 2 * PI + omega2;

  // translate chart into c[abc][xyz]
  cax = 0;
  cay = 0;
  cbx = cat[cb].x + transx - orgx;
  cby = cat[cb].y + transy - orgy;
  ccx = cat[cc].x + transx - orgx;
  ccy = cat[cc].y + transy - orgy;

  pax = 0;
  pay = 0;
  pbx = pic.s[pb].x - orgx;
  pby = pic.s[pb].y - orgy;
  pcx = pic.s[pc].x - orgx;
  pcy = pic.s[pc].y - orgy;

  // rotate chart into c[abc]r[xyz] and the picture int p[abc]r[xyz]
  // rotation is done around the common A/A' point

  carx = cax;
  cary = cay;

  omegac = atan2(cby,cbx);
  if (omegac<0)
    omegac = 2 * PI + omegac;

  cbrx = sqrt(cbx*cbx + cby*cby) * cos(omegac - omega2);
  cbry = sqrt(cbx*cbx + cby*cby) * sin(omegac - omega2);

  omegac = atan2(ccy,ccx);
  if (omegac<0)
    omegac = 2 * PI + omegac;

  ccrx = sqrt(ccx*ccx + ccy*ccy) * cos(omegac - omega2);
  ccry = sqrt(ccx*ccx + ccy*ccy) * sin(omegac - omega2);

  parx = 0;
  pary = 0;
  pbrx =   sqrt(pbx*pbx + pby*pby);
  pbry =   0;

  omegac = atan2(pcy,pcx);
  if (omegac<0)
    omegac = 2 * PI + omegac;

  pcrx = sqrt(pcx*pcx + pcy*pcy) * cos(omegac - omega1);
  pcry = sqrt(pcx*pcx + pcy*pcy) * sin(omegac - omega1);

  //  printf("  A: %6.1f %6.1f    %6.1f %6.1f    %6.1f %6.1f\n",
  //	 parx, pary, cax, cay, carx, cary);
  //  printf("  B: %6.1f %6.1f    %6.1f %6.1f    %6.1f %6.1f\n",
  //	 pbrx, pbry, cbx, cby, cbrx, cbry);
  //  printf("  C: %6.1f %6.1f    %6.1f %6.1f    %6.1f %6.1f\n",
  //	 pcrx, pcry, ccx, ccy, ccrx, ccry);

  magx = pbrx / cbrx; // not used anymore, except for flip;
  magy = pcry / ccry; // a full linear transform is needed

}

// transform coordinates from the plane of the chart into the plane
// of the picture

static void chart_to_pic(double x, double y, double catfact,
		  double *newx, double *newy, int with_ltr) {
  double xx, yy;
  double x0, y0;
  double omegac;
  double dd;

  xx = x + transx - orgx;
  yy = y + transy - orgy;
  omegac = atan2(yy,xx);
  if(omegac<0)
    omegac = 2 * PI + omegac;

  dd = sqrt(xx*xx + yy*yy);
  x0  = dd * cos(omegac - omega2);
  y0  = dd * sin(omegac - omega2);
  //  xx = x0 * magx;
  //  yy = y0 * magy;
  xx = x0;
  if (magy<0)
    yy = - y0;
  else
    yy = y0;

  omegac = atan2(yy,xx);
  if(omegac<0)
    omegac = 2 * PI + omegac;

  dd = sqrt(xx*xx + yy*yy);
  x0 = dd * cos(omegac  + omega1);
  y0 = dd * sin(omegac  + omega1);

  if (with_ltr) {
    *newx = (atr * x0 + btr * y0) + orgx;
    *newy = (ctr * x0 + dtr * y0) + orgy;
  } else {
    *newx = x0 + orgx;
    *newy = y0 + orgy;
  }
  
}

static int magcomp(const void *a, const void *b) {
  if (((STAR*)a)->mag > ((STAR*)b)->mag)
    return 1;
  if (((STAR*)a)->mag == ((STAR*)b)->mag)
    return 0;
  return -1;
}


#define MINPERIM 60
#define MINANGLE 0.2

static TRI tri(int i, int j, int k, STAR *s) { // compute a triangle from 3 stars
  // modified to exclude very small, isoscele or very obtuse triangles
  // effect of these limitations with various tolerances on perimeter, angles, etc
  // is not taken into account

  TRI r;
  double a,b,c;

  r.sa = i;
  r.sb = j;
  r.sc = k;
  
  a = r.a = dist(s[j].x, s[j].y, s[k].x, s[k].y);
  b = r.b = dist(s[i].x, s[i].y, s[k].x, s[k].y);
  c = r.c = dist(s[i].x, s[i].y, s[j].x, s[j].y);

  r.al = acos((a*a - c*c + b*b) / (2*a*b));
  r.be = acos((c*c - a*a + b*b) / (2*c*b));
  r.ga = (3.1415927) - r.al - r.be;
  
  r.perim = a+b+c;

  r.useful = 1;

  if ((a>=maxdist) || (b>=maxdist) || (c>=maxdist)) {
    r.useful = 0;
  }
  /*  if (r.al<MINANGLE || r.be<MINANGLE || r.ga<MINANGLE)
    r.useful = 0;

  if (r.perim<MINPERIM)
  r.useful = 0;*/
  return r;
}

static int tri_alike(TRI pi, TRI ca, double per_epsilon, double delta) {
  // compare two triangles for similarity; allow a tolerance of per_epsilon
  // for the length of the triangle perimeter and a of delta for the angles
  // pi is presumably in the picture, ca is presumably in the catalog
  // if the two triangles are judged alike, a positive is returned, otherwise 0
  // the positive indicates the relationship between vertices

  if (pi.perim > ca.perim) {
    if (1.0 - (ca.perim/pi.perim) > per_epsilon) {
      return 0;
    }
  }
  else {
    if (1.0 - (pi.perim/ca.perim) > per_epsilon) {
      return 0;
    }
  }

  // perimeters are alike, let's check the angles

  if ( (fabs(pi.al - ca.al) < delta) &&
       (fabs(pi.be - ca.be) < delta) &&
       (fabs(pi.ga - ca.ga) < delta))
    return 1;

  if ( (fabs(pi.al - ca.be) < delta) &&
       (fabs(pi.be - ca.ga) < delta) &&
       (fabs(pi.ga - ca.al) < delta))
    return 2;

  if ( (fabs(pi.al - ca.ga) < delta) &&
       (fabs(pi.be - ca.al) < delta) &&
       (fabs(pi.ga - ca.be) < delta))
    return 3;

  if ( (fabs(pi.al - ca.al) < delta) &&
       (fabs(pi.be - ca.ga) < delta) &&
       (fabs(pi.ga - ca.be) < delta))
    return 4;

  if ( (fabs(pi.al - ca.be) < delta) &&
       (fabs(pi.be - ca.al) < delta) &&
       (fabs(pi.ga - ca.ga) < delta))
    return 5;

  if ( (fabs(pi.al - ca.ga) < delta) &&
       (fabs(pi.be - ca.be) < delta) &&
       (fabs(pi.ga - ca.al) < delta))
    return 6;

  return 0;
}

static int init_pic(float *xs, float *ys, float *ms, int nismax) {
  // build a picture description from xs and ys, complete with triangles

  int i,j,k,nt,l;
  int maxtri;
  int nin[nis];

  for (i=0; i<nis; i++)
    nin[i] = 0;

  pic.s = malloc(nis * sizeof(STAR));
  pic.t = malloc((maxtri = selective_compute_ntri(nismax)) * sizeof(TRI));

  pic.ntri = 0;
  pic.nstar = nis;

  if (!pic.s || !pic.t) {
    error("Too many combinations\n");
    return (-1); // error
  }

  for(i=0; i<nis; i++) {
    pic.s[i].x = xs[i];
    pic.s[i].y = ys[i];
    pic.s[i].mag = ms[i];
    pic.s[i].elt = i;
    strcpy(pic.s[i].id, "pppppnnnnn");
  }

  qsort(pic.s,nis,sizeof(STAR),magcomp);

  pic.ntri = 0;
  for(i=2; i<nismax; i++) {
    for(j=1; j<i; j++) {
      for(k=0; k<j; k++) {
	if ((nin[i]>3) || (nin[j]>3) || nin[k]>3)
	  continue;
	pic.t[pic.ntri] = tri(i,j,k,pic.s); // assumed to always lead to a useful triangle
	if (pic.t[pic.ntri].useful) {
	  pic.ntri ++;
	  nin[i]++; nin[j]++; nin[k]++;
  	}
	//	if ((pic.ntri>=1000) || (pic.ntri>=maxtri)) // maximum number of triangles to consider in the picture
	//	  break; 
      }
    }
  }
  return 0; // ok
}

static int verbosef = 0;

static void assess_triangle_header() {
  printf(" PA  PB  PC   CA  CB  CC  OMEGA  DX     DY     DMAG MERR\n");
}

static void assess_triangle_hypothesis(int pa, int pb, int pc, // picture stars
				int ca, int cb, int cc, // catalog stars
				double catfact, double tolerance, int match) {

  double difest,magerr,omega;
  double nx0,ny0,nx1,ny1;

  compute_transform(pa,pb,pc,ca,cb,cc);
  chart_to_pic(cat[0].x,cat[0].y,1.0,&nx0,&ny0,0);
  chart_to_pic(cat[10].x,cat[10].y,1.0,&nx1,&ny1,0); // ???
  difest = ((pic.s[pa].mag + pic.s[pb].mag + pic.s[pc].mag) / 3.0) -
    ((cat[ca].mag + cat[cb].mag + cat[cc].mag) / 3.0);
  magerr = ((pic.s[pa].mag - cat[ca].mag - difest) *
            (pic.s[pa].mag - cat[ca].mag - difest)  +
            (pic.s[pb].mag - cat[cb].mag - difest) *
            (pic.s[pb].mag - cat[cb].mag - difest)  +
            (pic.s[pc].mag - cat[cc].mag - difest) *
            (pic.s[pc].mag - cat[cc].mag - difest)) / 3.0;
  magerr = sqrt(magerr);
  omega = omega1-omega2;
  if (omega > (2.0*PI))
    omega = omega - 2.0*PI;
  if (omega < 0.0)
    omega = omega + 2.0*PI;
  //  if (magerr<0.2) {
  if (list_triangle_matches_f) {
    printf("%3d %3d %3d  %3d %3d %3d %6.2f %6.1f,%6.1f %6.1f,%6.1f %4.1f %6.4f\n",
	   pa,pb,pc,ca,cb,cc,omega,nx0,ny0,difest,magerr);
  }
  if (tsfill<TSMAX) {
    ts[tsfill].pa = pa;
    ts[tsfill].pb = pb;
    ts[tsfill].pc = pc;
    ts[tsfill].ca = ca;
    ts[tsfill].cb = cb;
    ts[tsfill].cc = cc;
    ts[tsfill].omega = omega;
    ts[tsfill].dx0 = nx0;
    ts[tsfill].dy0 = ny0;
    ts[tsfill].dx1 = nx1;
    ts[tsfill].dy1 = ny1;
    ts[tsfill].dmag = difest;
    ts[tsfill].merr = magerr;
    ts[tsfill].repeated = 0;
    tsfill ++;
  }
  //    test_triangle_hypothesis(pa,pb,pc,ca,cb,cc,catfact,tolerance,match);
  //  }
}

static float match_distance(TRISOL a, TRISOL b) {
  return sqrt(((a.omega - b.omega) * (a.omega - b.omega) / 0.01 +
	       (a.dx0 - b.dx0) * (a.dx0 - b.dx0) / 400.0 +
	       (a.dy0 - b.dy0) * (a.dy0 - b.dy0) / 400.0 +
	       (a.dx1 - b.dx1) * (a.dx1 - b.dx1) / 400.0 +
	       (a.dy1 - b.dy1) * (a.dy1 - b.dy1) / 400.0 +
	       (a.dmag - b.dmag) * (a.dmag - b.dmag) / 0.25) / 6.0);
}


static float idx, idy, idmag, iomega;
static int insol;

static void locate_repeats() {
  int i,j;
  int pivotype, pivonum; // type and number of objects of a kind
  int pvn; // candidate pivonum

  pivotype = -1;
  pivonum = 0;

  if(list_repeated_solutions_f) {
    printf("REPEATED SOLUTIONS\n");
    printf(" PA  PB  PC   CA  CB  CC  OMEGA  DX0    DY0    DX1    DY1    DMAG MERR\n");
  }
  for(i=0; i<tsfill; i++)
    ts[i].like = -1; // unlike any other
  for(i=0; i<tsfill; i++) {
    if (!ts[i].repeated) {
      for(j=0; j<nis; j++)
	autosol[j] = -1;
      autosol[ts[i].pa] = ts[i].ca;
      autosol[ts[i].pb] = ts[i].cb;
      autosol[ts[i].pc] = ts[i].cc;
      for(j=0; j<tsfill; j++) {
	if (i==j)
	  continue;
	if (match_distance(ts[i],ts[j]) < 1.0) {
	  if(list_repeated_solutions_f) {
	    if (!ts[i].repeated) 
	      printf("\n%3d %3d %3d  %3d %3d %3d %6.2f %6.1f %6.1f %4.1f %6.4f\n",
		     ts[i].pa,ts[i].pb,ts[i].pc,ts[i].ca,ts[i].cb,ts[i].cc,
		     ts[i].omega,ts[i].dx0,ts[i].dy0,ts[i].dmag,ts[i].merr);
	    printf("%3d %3d %3d  %3d %3d %3d %6.2f %6.1f %6.1f %4.1f %6.4f\n",
		   ts[j].pa,ts[j].pb,ts[j].pc,ts[j].ca,ts[j].cb,ts[j].cc,
		   ts[j].omega,ts[j].dx0,ts[j].dy0,ts[j].dmag,ts[j].merr);
	  }
	  ts[i].repeated++;
	  ts[j].repeated++;
	  ts[i].like = ts[j].like = i;
	  if ((autosol[ts[j].pa]>=0) && (autosol[ts[j].pa]!=ts[j].ca))
	    autosol[ts[j].pa] = -2;
	  else
	    autosol[ts[j].pa] = ts[j].ca;
	  if ((autosol[ts[j].pb]>=0) && (autosol[ts[j].pb]!=ts[j].cb))
	    autosol[ts[j].pb] = -2;
	  else
	    autosol[ts[j].pb] = ts[j].cb;
	  if ((autosol[ts[j].pc]>=0) && (autosol[ts[j].pc]!=ts[j].cc))
	    autosol[ts[j].pc] = -2;
	  else
	    autosol[ts[j].pc] = ts[j].cc;
	}
      }
      for(pvn=0, j=0; j<nis; j++)
	if (autosol[j]!=-1)
	  pvn ++;
      if (pvn>pivonum) {
	pivotype = i;
	pivonum = pvn;
      }
    }
  }
  for(i=0; i<full_nis; i++)
    autosol[i] = -1;
  idx = idy = iomega = idmag = 0.0;
  insol = 0;
  for(i=0; i<tsfill; i++) {
    if(ts[i].like==pivotype) {
      if (list_final_solution_f)
	printf("\n%3d %3d %3d  %3d %3d %3d %6.2f %6.1f,%6.1f %6.1f,%6.1f %4.1f %6.4f\n",
	       ts[i].pa,ts[i].pb,ts[i].pc,ts[i].ca,ts[i].cb,ts[i].cc,
	       ts[i].omega,ts[i].dx0,ts[i].dy0,ts[i].dx1,ts[i].dy1,ts[i].dmag,ts[i].merr);
      idx += ts[i].dx0;
      idy += ts[i].dy0;
      iomega += ts[i].omega;
      idmag += ts[i].dmag;
      insol ++;
      if ((autosol[ts[i].pa]<0) || (autosol[ts[i].pa]>ts[i].ca)) // just take the brightest
	autosol[ts[i].pa] = ts[i].ca;
      if ((autosol[ts[i].pb]<0) || (autosol[ts[i].pb]>ts[i].cb)) // just take the brightest
	autosol[ts[i].pb] = ts[i].cb;
      if ((autosol[ts[i].pc]<0) || (autosol[ts[i].pc]>ts[i].cc)) // just take the brightest
	autosol[ts[i].pc] = ts[i].cc;
    }
  }

  if (insol) {
    idx /= insol;
    idy /= insol;
    iomega /= insol;
    idmag /= insol;
  }

  compute_transform(ts[pivotype].pa,ts[pivotype].pb,ts[pivotype].pc,
		    ts[pivotype].ca,ts[pivotype].cb,ts[pivotype].cc); // and leave it that way
  ref_solution = pivotype;
  //  depict_solution(0,im,ts[pivotype].dmag);

  if(list_final_solution_f) {
    printf("SOLUTION: \n");
    for(i=0; i<nis; i++) {
      if (autosol[i]!=-1)
	printf("%2d: %3d=GSC-%s\n", i, autosol[i], cat[autosol[i]].id);
    }
  }
}

static void list_solutions() {
  int i;

  printf("REPEATED SOLUTIONS\n");
  printf(" PA  PB  PC   CA  CB  CC  OMEGA  DX     DY     DMAG MERR\n");
  for(i=0; i<tsfill; i++)
    if (ts[i].repeated)
      printf("%3d %3d %3d  %3d %3d %3d %6.2f %6.1f %6.1f %4.1f %6.4f\n",
	 ts[i].pa,ts[i].pb,ts[i].pc,ts[i].ca,ts[i].cb,ts[i].cc,
	     ts[i].omega,ts[i].dx0,ts[i].dy0,ts[i].dmag,ts[i].merr);
      
}

static  float len_tol_min = 0.85;
static  float len_tol_max = 1.15;
static  float mag_tol_min = -1.5;
static  float mag_tol_max = +1.5;

static void selective_find_alikes() {

  int i,j,k,l,al,trues;
  TRI ca;
  double cfact;

  float magdif, magx;

  float pic_AB, pic_AC, pic_BC;
  float cat_AB, cat_AC, cat_BC, cat_AX;
  int cata, catb, catc, catx;

  float hyptol = 5.0;

  int candidate_B[catsz], candidate_C[catsz];

  int first_B, last_B, first_C, last_C; // to limit 

  for (l=0; l<pic.ntri; l++) {
    pic_AB = dist(pic.s[pic.t[l].sa].x, pic.s[pic.t[l].sa].y,
		  pic.s[pic.t[l].sb].x, pic.s[pic.t[l].sb].y);
    pic_AC = dist(pic.s[pic.t[l].sa].x, pic.s[pic.t[l].sa].y,
		  pic.s[pic.t[l].sc].x, pic.s[pic.t[l].sc].y);
    pic_BC = dist(pic.s[pic.t[l].sb].x, pic.s[pic.t[l].sb].y,
		  pic.s[pic.t[l].sc].x, pic.s[pic.t[l].sc].y);

    // consider star A to be each of the catalog stars in part
    for (cata=0; cata<catsz; cata++) {
      first_B = last_B = first_C = last_C = -1;
      magdif = pic.s[pic.t[l].sa].mag - cat[cata].mag;
      for (catx=0; catx<catsz; catx++) { // check if it is a suitable B or C star
	if (catx==cata) {
	  candidate_B[catx] = candidate_C[catx] = 0;
	  continue;
	}
	magx = magdif + cat[catx].mag;

	if (candidate_B[catx] = ((pic.s[pic.t[l].sb].mag > (magx + mag_tol_min)) &&
	    (pic.s[pic.t[l].sb].mag < (magx + mag_tol_max)))) {
	  if (first_B==-1)
	    first_B = catx;
	} else {
	  if ((first_B!=-1) && (last_B==-1))
	    last_B = catx;
	}

	if (candidate_C[catx] = ((pic.s[pic.t[l].sc].mag > (magx + mag_tol_min)) &&
	    (pic.s[pic.t[l].sc].mag < (magx + mag_tol_max)))) {
	  if (first_C==-1)
	    first_C = catx;
	} else {
	  if ((first_C!=-1) && (last_C==-1))
	    last_C = catx;
	}

	cat_AX = dist(cat[cata].x,cat[cata].y,cat[catx].x,cat[catx].y);

	if (candidate_B[catx]) { // test also that distances are reasonable
	  candidate_B[catx] = (cat_AX > (pic_AB * len_tol_min)) &&
	    (cat_AX < (pic_AB * len_tol_max));
	}

	if (candidate_C[catx]) {
	  candidate_C[catx] = (cat_AX > (pic_AC * len_tol_min)) &&
	    (cat_AX < (pic_AC * len_tol_max));
	}
      }

      if ((first_B!=-1) && (last_B)==-1)
	last_B = catsz;
      if ((first_C!=-1) && (last_C)==-1)
	last_C = catsz;

      // check triangles cata,catb,catc for which magnitudes and distances
      // are within reasonable range, for angles within reasonable range

      if (((first_B==-1) && (last_B==-1)) ||
	  ((first_C==-1) && (last_C==-1)))
	continue; // the hypothesis pic.s[l].sa is cat[cata] is rejected
                  // on the grounds that magnitudes of B or C would be outside the catalog

      if ((first_B!=-1) && (last_B==-1))
	last_B = catsz;
      if ((first_C!=-1) && (last_C!=-1))
	last_C = catsz;

      for (catb=first_B; catb<last_B; catb++)
	for (catc=first_C; catc<last_C; catc++) {
	  if(candidate_B[catb] && candidate_C[catc] && (catb!=catc)) {	    
	    ca = tri(cata,catb,catc,cat);
	    trues = (cata==handsol[pic.t[l].sa]) && (catb==handsol[pic.t[l].sb]) &&
	      (catc==handsol[pic.t[l].sc]); 
	    al = tri_alike(pic.t[l],ca,0.15,0.01); // 0.05 0.001
	    if (trues && list_triangle_matches_f) {
	      printf("Read solution %d=%s, %d=%s, %d=%s, al=%d\n",
		     pic.t[l].sa, cat[cata].id, pic.t[l].sb, cat[catb].id,
		     pic.t[l].sc, cat[catc].id, al);
	    }
	    if (al==1) {
	      cfact = pic.t[l].perim / ca.perim;
	      assess_triangle_hypothesis(pic.t[l].sa, pic.t[l].sb, pic.t[l].sc,
					 ca.sa, ca.sb, ca.sc, cfact, hyptol,1);
	    }
	  }
	}
    }
  }
}

static float astromerr(float dx, float dy, float omega, float amp) {
  int i;
  float err,lerr;
  float dd,th,catx,caty;

  err = 0.0;
  for(i=1; i<nis; i++) {
    if(autosol[i]!=-1) {
      dd = sqrt(cat[autosol[i]].x*cat[autosol[i]].x + cat[autosol[i]].y*cat[autosol[i]].y)
	* amp;
      th = atan2(cat[autosol[i]].y,cat[autosol[i]].x) + omega;
      catx = dx + dd*cos(th);
      caty = dy + dd*sin(th);
      lerr = (catx - pic.s[i].x)*(catx - pic.s[i].x) +
	(caty - pic.s[i].y)*(caty - pic.s[i].y);
      //      printf("lerr[%d]=%f\n",i,lerr);
      if(lerr<3000.0)
	err += lerr;
    }
  }
  return err;
}

static optimize(float *dx, float *dy, float *omega, float *amp) {
  int i;
  float err = 0.0;
  float nerr;
  float dxstep=0.1;
  float dystep=0.1;
  float ampstep=0.001;
  float omegastep=0.00001;

  err = astromerr(*dx,*dy,*omega,*amp);
  for(i=0; i<1000; i++) {
    //    printf("err=%f dx=%f dy=%f omega=%f amp=%f\n",err,*dx,*dy,*omega,*amp);
    if(err > astromerr(*dx-dxstep,*dy,*omega,*amp)) {
      *dx -= dxstep;
    } else {
      if(err > astromerr(*dx+dxstep,*dy,*omega,*amp))
	*dx += dxstep;
    }
    
    if(err > astromerr(*dx,*dy-dystep,*omega,*amp)) {
      *dy -= dystep;
    } else {
      if(err > astromerr(*dx,*dy+dystep,*omega,*amp))
	*dy += dystep;
    }
    
    if(err > astromerr(*dx,*dy,*omega,*amp+ampstep)) {
      *amp += ampstep;
    } else {
      if(err > astromerr(*dx+dxstep,*dy,*omega,*amp-ampstep))
	*amp -= ampstep;
    }

    if(err > astromerr(*dx,*dy,*omega+omegastep,*amp)) {
      *omega += omegastep;
    } else {
      if(err > astromerr(*dx+dxstep,*dy,*omega,*omega-omegastep))
	*omega -= omegastep;
    }

    nerr = astromerr(*dx,*dy,*omega,*amp);
    if ((err-nerr) < 0.1) // arbitrary limit
      break;
    err = nerr;
  }
}

void set_amplif_tolerance(float mi, float ma) {
  len_tol_min = mi;
  len_tol_max = ma;
}

void set_mag_tolerance(float mi, float ma) {
  mag_tol_min = mi;
  mag_tol_max = ma;
}

int starmatch(float *sx, float *sy, float *sm, int sn,    // input sources in picture
              float *cx, float *cy, float *cm, int cn,    // input catalog
              int *cix, float *omega, float *dx, float *dy,
              float *ampx, float *ampy, float *dmag)

{
  int i;
  int nreps=0;
  float dxs,dys,dd,th,ddd;

  full_nis = nis = sn;
  cat = (STAR *)malloc(cn * sizeof(STAR));

  // block of assignments to ensure that previous state does not
  // influence current calls
  transx=0.0;
  transy=0.0;
  omega2=0.0;
  orgx=0.0;
  orgy=0.0;
  magx=1.0;
  magy=1.0;
  omega1=0.0;
  atr=1.0;
  btr=0.0;
  ctr=0.0;
  dtr=1.0;
  // but I guess this was it...
  tsfill = 0;

  if (cat==NULL) {
    fprintf(stderr,"Could not allocated %d star places\n",cn);
    return -100;
  }
  for (i=0; i<cn; i++) {
    cat[i].x = cx[i];
    cat[i].y = cy[i];
    cat[i].mag = cm[i];
    cat[i].elt = i;
  }
  catsz = cn;
  qsort(cat,cn,sizeof(STAR),magcomp);

  if (catsz>100)
    catsz = 100;

  if (init_pic(sx, sy, sm, sn))
    return -101; // could not alloc or something
  selective_find_alikes();
  locate_repeats();

  dxs = dys = 0.0;
  for(i=0; i<nis; i++) { // unsorted arguments are supported
    if (autosol[i]!=-1) {
      dd = sqrt(cat[autosol[i]].x*cat[autosol[i]].x + cat[autosol[i]].y*cat[autosol[i]].y);
      th = atan2(cat[autosol[i]].y, cat[autosol[i]].x) + iomega;
      dxs += - dd*cos(th) + pic.s[i].x;
      dys += - dd*sin(th) + pic.s[i].y;
      cix[pic.s[i].elt] = cat[autosol[i]].elt;
      nreps ++;
    } else
      cix[pic.s[i].elt] = -1;
  }
  /*
  for(i=0; i<nis; i++) { // eliminate mis-associations (which are rare)
    if (autosol[i]!=-1) {
      dd = sqrt(cat[autosol[i]].x*cat[autosol[i]].x + cat[autosol[i]].y*cat[autosol[i]].y);
      th = atan2(cat[autosol[i]].y, cat[autosol[i]].x) + iomega;
      ddd = sqrt((pic.s[i].x - dd*cos(th)- dxs/nreps)*(pic.s[i].x - dd*cos(th)- dxs/nreps) +
		 (pic.s[i].y - dd*sin(th)- dys/nreps)*(pic.s[i].y - dd*sin(th)- dys/nreps));
      //      printf("%f\n",ddd);
      if(ddd > 50.0) { // max limit to consider the association correct 
	dxs -= - dd*cos(th) + pic.s[i].x;
	dys -= - dd*sin(th) + pic.s[i].y;
	autosol[i] = -1;
	cix[pic.s[i].elt] = -1;
	nreps --;
      }
    }
    }*/

  if (pic.s!=NULL) free(pic.s);
  if (pic.t!=NULL) free(pic.t);
  if (cat!=NULL) free(cat);
  *omega = iomega;
  *dx = dxs/nreps;
  *dy = dys/nreps;
  *ampx = 1.0;
  *ampy = 1.0;
  *dmag = idmag;
  optimize(dx,dy,omega,ampx);
  *ampy = *ampx;
  return nreps;
}

